@@ -29,6 +29,8 @@ from posixpath import normpath
 PSL_URL = 'http://mxr.mozilla.org/mozilla-central/source/netwerk/dns/effective_tld_names.dat?raw=1'
 
 def _get_public_suffix_list():
+    """Get the public suffix list.
+    """
     local_psl = os.environ.get('PUBLIC_SUFFIX_LIST')
     if local_psl:
         psl_raw = open(local_psl).readlines()
@@ -44,10 +46,27 @@ def _get_public_suffix_list():
 PSL = _get_public_suffix_list()
 
 
-SCHEMES = ['http', 'https', 'ftp']
+SCHEMES = ['http', 'https', 'ftp', 'sftp', 'file', 'gopher', 'imap', 'mms',
+           'news', 'nntp', 'telnet', 'prospero', 'rsync', 'rtsp', 'rtspu',
+           'svn', 'git']
 
 def normalize(url):
+    """Normalize a URL
+    """
     parts = extract(url)
+    return _assemble(parts, default_path='/')
+
+def encode(url):
+    """
+    """
+    parts = extract(url)
+    idna = lambda x: x.decode('utf-8').encode('idna')
+    encoded = ParseResult(*(idna(p) for p in parts))
+    return _assemble(encoded)
+
+def _assemble(parts, default_path=''):
+    """
+    """
     nurl = ''
     if parts.scheme:
         if parts.scheme in SCHEMES:
@@ -56,6 +75,8 @@ def normalize(url):
             nurl += parts.scheme + ':'
     if parts.username and parts.password:
         nurl += parts.username + ':' + parts.password + '@'
+    elif parts.username:
+        nurl += parts.username + '@'
     if parts.subdomain:
         nurl += parts.subdomain + '.'
     nurl += parts.domain
@@ -64,19 +85,28 @@ def normalize(url):
     if parts.port and parts.port != '80':
         nurl += ':' + parts.port
     if parts.path:
-        nurl += normpath(parts.path)
+        nurl += _normalize_path(parts.path)
     elif parts.scheme in SCHEMES:
-        nurl += '/'
+        nurl += default_path
     if parts.query:
         nurl += '?' + parts.query
     if parts.fragment:
         nurl += '#' + parts.fragment
     return nurl
 
+def _normalize_path(path):
+    """
+    """
+    npath = normpath(path)
+    if path.endswith('/') and not npath.endswith('/'):
+        npath += '/'
+    return npath.replace('//', '/')
 
+
+IP_RE = re.compile(r'[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}')
 PORT_RE = re.compile(r'(?<=.:)[1-9]+[0-9]{0,4}$')
 SCHEME_RE = re.compile(r'^[a-zA-Z]+:(//)?')
-USER_RE = re.compile(r'^[^:@]*:[^:@]*@.*$')
+USER_RE = re.compile(r'^[^:@]*:?[^:@]+@.*$')
 SplitResult = namedtuple('SplitResult', ['scheme', 'netloc', 'path', 'query',
                                          'fragment'])
 ParseResult = namedtuple('ParseResult', ['scheme', 'username', 'password',
@@ -84,7 +114,7 @@ ParseResult = namedtuple('ParseResult', ['scheme', 'username', 'password',
                                          'path', 'query', 'fragment'])
 
 def split(url):
-    """Split URLs into scheme, netloc, path, query and fragment
+    """Split URL into scheme, netloc, path, query and fragment
     """
     scheme = netloc = path = query = fragment = ''
     if SCHEME_RE.search(url):
@@ -98,14 +128,21 @@ def split(url):
     l_frag = rest.find('#')
     if l_path > 0:
         netloc = rest[:l_path]
-        if l_query > 0:
+        if l_query > 0 and l_frag > 0:
+            path = rest[l_path:min(l_query, l_frag)]
+        elif l_query > 0:
             path = rest[l_path:l_query]
         elif l_frag > 0:
             path = rest[l_path:l_frag]
         else:
             path = rest[l_path:]
     else:
-        netloc = rest
+        if l_query > 0:
+            netloc = rest[:l_query]
+        elif l_frag > 0:
+            netloc = rest[:l_frag]
+        else:
+            netloc = rest
     if l_query > 0:
         if l_frag > 0:
             query = rest[l_query+1:l_frag]
@@ -119,16 +156,29 @@ def split(url):
     return SplitResult(scheme, netloc, path, query, fragment)
 
 def _clean_netloc(netloc):
+    """Remove trailing '.'s and ':'s from a URL and tolower
+    """
     return netloc.rstrip('.:').decode('utf-8').lower().encode('utf-8')
 
 def _split_netloc(netloc):
+    """Split netloc into username, password, subdomain, domain, tld and port
+    """
     username = password = subdomain = tld = port = ''
     if USER_RE.search(netloc):
         user_pw, netloc = netloc.split('@', 1)
-        username, password = user_pw.split(':', 1)
+        if ':' in user_pw:
+            username, password = user_pw.split(':', 1)
+        else:
+            username = user_pw
+    if IP_RE.search(netloc):
+        if PORT_RE.search(netloc):
+            domain, port = netloc.split(':')
+        else:
+            domain = netloc
+        return username, password, '', domain, '', port
     netloc = _clean_netloc(netloc)
     if '.' not in netloc:
-        return '', '', '', netloc, '', ''
+        return username, password, '', netloc, '', ''
     if PORT_RE.search(netloc):
         domain, port = netloc.split(':')
     else:
@@ -154,21 +204,24 @@ def _split_netloc(netloc):
     return username, password, subdomain, domain, tld, port
 
 def parse(url):
+    """Parse URL
+    """
     parts = split(url)
     if parts.scheme:
         netloc = parts.netloc
         (username, password, subdomain, domain, tld, port) = _split_netloc(netloc)
     else:
         username = password = subdomain = domain = tld = port = ''
-    path = parts.path if parts.path else ''
     return ParseResult(parts.scheme, username, password, subdomain, domain, tld,
-                       port, path, parts.query, parts.fragment)
+                       port, parts.path, parts.query, parts.fragment)
 
 def extract(url):
+    """
+    """
     parts = split(url)
     if parts.scheme:
         netloc = parts.netloc
-        path = parts.path if parts.path else ''
+        path = parts.path
     else:
         netloc = parts.path
         path = ''
